from .code import CodeSearchResult
from .commit import CommitSearchResult
from .issue import IssueSearchResult
from .repository import RepositorySearchResult
from .user import UserSearchResult


__all__ = (
    "CodeSearchResult",
    "CommitSearchResult",
    "IssueSearchResult",
    "RepositorySearchResult",
    "UserSearchResult",
)
